/* Copyright (C) 2022 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Volker Friese [committer] */

#ifndef ALGO_DETECTORS_STS_StsReadoutConfigLegacy_H
#define ALGO_DETECTORS_STS_StsReadoutConfigLegacy_H

#include <map>
#include <sstream>
#include <utility>
#include <vector>

namespace cbm::algo
{


  /** @class StsReadoutConfigLegacy
   ** @author Volker Friese <v.friese@gsi.de>
   ** @since 3 March 2022
   ** @brief Provides the hardware-to-software address mapping for the CBM-STS
   **
   ** The hardware address as provided in the raw data stream is specified in terms of the
   ** equipment identifier (specific to one FLES component) and the elink number with in
   ** component. This is to be translated into the module address and the ASIC number within
   ** the module.
   ** The mapping of the two address spaces is hard-coded in this class.
   **
   ** NOTE: This class is deprecated. It's replaced by ReadoutConfig in detector/sts. The new class no longer hardcodes the mapping. It can be read from a configuration file instead.
   **/

  class StsReadoutConfigLegacy {

  public:
    /** @brief Constructor **/
    StsReadoutConfigLegacy();


    /** @brief Destructor **/
    virtual ~StsReadoutConfigLegacy();


    /** @brief Equipment in the configuration
     ** @return Vector of equipment IDs
     **/
    std::vector<uint16_t> GetEquipmentIds();


    /** @brief Number of elinks of a component
     ** @param Equipment ID
     ** @return Number of elinks
     **/
    size_t GetNumElinks(uint16_t equipmentId);


    /** @brief Total number of elinks for STS
     ** @return Number of elinks
     **/
    size_t GetNumElinks();


    /** @brief API: Mapping from component and elink to address and ASIC number
     ** @param equipId     Equipment identifier (component)
     ** @param elink       Elink number within component
     ** @return (module address, ASIC number within module)
     */
    std::pair<int32_t, uint16_t> Map(uint16_t equipId, uint16_t elink);


    /** @brief API: Mapping from module address and ASIC number to walk coefficients. 
     ** @param modAddress     module address
     ** @param asic           ASIC number within module
     ** @return std::vector<double> of walk coefficients
     */
    std::vector<double> WalkMap(int32_t modAddress, uint16_t asic);


    /** @brief Debug output of readout map **/
    std::string PrintReadoutMap();


  private:
    // --- STS readout map
    // --- Map index: (equipment, elink), map value: (module address, ASIC number in module)
    std::map<uint16_t, std::vector<std::pair<int32_t, uint16_t>>> fReadoutMap = {};  //!

    // --- STS walk correction map
    // --- Map index: (module address, ASIC number in module, ADC value), map value: (walk coefficient)
    std::map<int32_t, std::vector<std::vector<double>>> fWalkMap;

    /** @brief Initialisation of readout map **/
    void Init();

    /** @brief Initialisation of walk map **/
    void InitWalkCoeffs();
  };

} /* namespace cbm::algo */

#endif /* ALGO_DETECTORS_STS_StsReadoutConfigLegacy_H_ */
